// This program is free software: you can use, modify and/or redistribute it
// under the terms of the simplified BSD License. You should have received a
// copy of this license along this program. If not, see
// <http://www.opensource.org/licenses/bsd-license.html>.
//
// Copyright (C) 2017, Javier Sánchez Pérez <jsanchez@ulpgc.es>
// All rights reserved.


#include "midway_video_equalization.h"
#include "utils.h"

#include <sys/time.h> 
#include <stdio.h> 
#include <stdlib.h> 
#include <string.h> 

#define PAR_DEFAULT_METHOD PIXEL_SORTING_METHOD
#define PAR_DEFAULT_SIGMA 30.0
#define PAR_DEFAULT_OUTNAME "output_video.raw"

/**
 *
 *  Print a help message 
 *
 */
void print_help(char *name) 
{
  printf("\n  Usage: %s raw_input_video width height nframes [OPTIONS] \n\n",
          name);
  printf("  This program calculates the Midway Equalization for a video.\n");
  printf("  'raw_input_video' is a video file in raw format (rgb24).\n");
  printf("  'width' is the width of the images in pixels.\n");
  printf("  'height' is the height of the images in pixels.\n");
  printf("  'nframes' is the number of frames in the video.\n");
  printf("  -----------------------------------------------\n");
  printf("  Converting to raw data:\n");
  printf("  'avconv -i video.mp4 -f rawvideo -pix_fmt rgb24 -y "
         "raw_video.raw'\n");
  printf("  to convert an mp4 video to raw format.\n");
  printf("  'avconv -f rawvideo -pix_fmt rgb24 -video_size 640x360 "
         "-framerate\n");
  printf("  30 -i output_video.raw -pix_fmt yuv420p output_video.mp4'\n");
  printf("  to convert a raw video to mp4 format.\n");
  printf("  -----------------------------------------------\n");
  printf("  More information in http://www.ipol.im \n\n");
  printf("  OPTIONS:\n");
  printf("  --------\n");
  printf("   -o name  output video name to write the computed raw video,\n");
  printf("              default value '%s'\n", PAR_DEFAULT_OUTNAME);
  printf("   -m N     equalization method,\n");
  printf("              0.histogram inversion; 1.image sorting;\n");
  printf("              default value %d\n", PAR_DEFAULT_METHOD);
  printf("   -s N     standard deviation for the Gaussian convolutions,\n");
  printf("              default value %f\n", PAR_DEFAULT_SIGMA);
  printf("   -d       use the dithering method. \n");
  printf("   -v       switch on verbose mode. \n\n\n");
}


/**
 *
 *  Read command line parameters 
 *
 */
int read_parameters(
  int   argc, 
  char  *argv[], 
  char  **input_video,
  char  *output_video,
  int   &method,
  float &sigma,
  int   &dithering,
  int   &nx,
  int   &ny,
  int   &nframes,
  int   &verbose
)
{
  if (argc < 5){
    print_help(argv[0]);
    return 0;
  }
  else{
    int i=1;
    
    //assign default values to the parameters
    strcpy(output_video, PAR_DEFAULT_OUTNAME);
    method=PAR_DEFAULT_METHOD;
    sigma=PAR_DEFAULT_SIGMA;
    dithering=0;
    verbose=0;
    
    //read parameters
    *input_video=argv[i]; i++;
    nx=atoi(argv[i]); i++;
    ny=atoi(argv[i]); i++;
    nframes=atoi(argv[i]); i++;;
    
    //read optional parameters from the command line
    while(i<argc)
    {
      if(strcmp(argv[i],"-o")==0)
        if(i<argc-1)
          strcpy(output_video, argv[++i]);

      if(strcmp(argv[i],"-m")==0)
        if(i<argc-1)
          method=atoi(argv[++i]);

      if(strcmp(argv[i],"-s")==0)
        if(i<argc-1)
          sigma=atof(argv[++i]);

      if(strcmp(argv[i],"-d")==0)
        dithering=1;

      if(strcmp(argv[i],"-v")==0)
        verbose=1;
      
      i++;
    }

    //check parameter values
    if(nx<=0 || ny<=0 || nframes<=0)
    {
      fprintf(stderr, "Error: Incorrect size of the input video.\n");
      exit(EXIT_FAILURE);
    }
    if(method<0 || method>1) method=PAR_DEFAULT_METHOD;
    if(sigma<=0) sigma=PAR_DEFAULT_SIGMA;
  }

  return 1;
}

/**
 *
 *  Main program:
 *  This program implements the midway equalization algorithm for a video
 *
 */
int main (int argc, char *argv[])
{
  int nx, ny, nz=3, nframes;
  int dithering, verbose, method;
  float sigma;
  char *input_video, output_video[200];

  //read the parameters from the console
  int result=read_parameters(
        argc, argv, &input_video, output_video, method,
        sigma, dithering, nx, ny, nframes, verbose
      );

  if(result)
  { 
    struct timeval start, end;

    if(verbose) 
      gettimeofday(&start, NULL);

    if(verbose)
      printf(
        " Frame width: %d, height: %d, number of frames: %d\n"
        " method: %d, sigma: %f, dithering: %d \n", 
        nx, ny, nframes, method, sigma, dithering
      );
    
    int size=nx*ny*nz*nframes;
    unsigned char *I=new unsigned char[size];
   
    size_t r=read_video(input_video, I, size);
    
    if(r<=0)
    {
      fprintf(stderr, "Error: Cannot read the input video '%s'.\n",
              input_video);
      return EXIT_FAILURE;
    }


    if(verbose) 
      printf(" Size of video in bytes %d\n", (int) r);

    //midway equalization
    midway_video_equalization(
      I, nx, ny, nz, nframes, method,
      sigma, dithering, verbose
    );

    write_video(output_video, I, size);
    
    if(verbose) 
    {
      gettimeofday(&end, NULL);
      float delay=((end.tv_sec-start.tv_sec)* 1000000u + 
                    end.tv_usec - start.tv_usec) / 1.e6; 
      printf("\n Time: %fs\n", delay);
    }
    
    delete []I;
    return EXIT_SUCCESS;
  }
  return EXIT_FAILURE;
}
